# -*- coding: utf-8 -*-
import numpy as np
import torch

import pandas as pd
from logging import getLogger
import os

from accelbrainbase.extractabledata.labeled_csv_extractor import LabeledCSVExtractor
from accelbrainbase.iteratabledata.ssda_transformer_iterator import SSDATransformerIterator as _SSDATransformerIterator
from accelbrainbase.noiseable_data import NoiseableData


class LabeledCSVSSDATransformerIterator(_SSDATransformerIterator):
    '''
    Iterator that draws from CSV files and generates `torch.Tensor` of labeled samples.
    '''

    def __init__(
        self,
        labeled_csv_extractor,
        train_csv_path,
        test_csv_path,
        target_csv_path=None,
        epochs=300,
        batch_size=20,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        ctx="cpu"
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            train_csv_path:                 `str` of path to CSV file in training.
            test_csv_path:                  `str` of path to CSV file in test.

            epochs:                         `int` of epochs of Mini-batch.
            batch_size:                     `int` of batch size of Mini-batch.
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
        '''
        if isinstance(labeled_csv_extractor, LabeledCSVExtractor) is False:
            raise TypeError("The type of `labeled_csv_extractor` must be `LabeledCSVExtractor`.")
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        if target_csv_path is None:
            target_csv_path = test_csv_path

        train_observed_arr, train_label_arr = labeled_csv_extractor.extract(train_csv_path)
        test_observed_arr, test_label_arr = labeled_csv_extractor.extract(test_csv_path)
        target_observed_arr, _ = labeled_csv_extractor.extract(target_csv_path)

        dataset_size = train_observed_arr.shape[0]
        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        train_label_df = pd.DataFrame(train_label_arr, columns=["tmp"])
        self.__label_n = train_label_df.tmp.drop_duplicates().values.shape[0]

        self.__train_observed_arr = train_observed_arr
        self.__train_label_arr = train_label_arr
        self.__test_observed_arr = test_observed_arr
        self.__test_label_arr = test_label_arr
        self.__target_observed_arr = target_observed_arr

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data

        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - encoder's observed data points in training.
            - decoder's observed data points in training.
            - encoder's masked data points in training.
            - decoder's masked data points in training.
            - encoder's observed data points in test.
            - decoder's observed data points in test.
            - encoder's masked data points in test.
            - decoder's masked data points in test.
            - Labeled data for downstream task in training.
            - Labeled data for downstream task in test.
            - `list` of encoder's observed data points in pretext task.
            - `list` of decoder's observed data points in pretext task.
            - `list` of encoder's masked data points in pretext task.
            - `list` of decoder's masked data points in pretext task.
            - `list` of labeled data in pretext task.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None
            training_label_arr, test_label_arr = None, None
            row_arr = np.arange(self.__train_observed_arr.shape[0])
            np.random.shuffle(row_arr)

            training_batch_arr = self.__train_observed_arr[row_arr[:self.batch_size]]
            training_batch_arr = torch.from_numpy(training_batch_arr)
            training_batch_arr = training_batch_arr.to(self.__ctx).float()
            training_batch_arr = self.pre_normalize(training_batch_arr)

            label_key_arr = self.__train_label_arr[row_arr[:self.batch_size]]
            label_key_arr = torch.from_numpy(label_key_arr)
            label_key_arr = label_key_arr.to(self.__ctx).float()
            training_label_arr = torch.one_hot(label_key_arr, self.__label_n)

            test_row_arr = np.arange(self.__test_observed_arr.shape[0])
            np.random.shuffle(test_row_arr)

            test_batch_arr = self.__test_observed_arr[test_row_arr[:self.batch_size]]
            test_batch_arr = torch.from_numpy(test_batch_arr)
            test_batch_arr = test_batch_arr.to(self.__ctx).float()
            test_batch_arr = self.pre_normalize(test_batch_arr)

            test_label_key_arr = self.__test_label_arr[test_row_arr[:self.batch_size]]
            test_label_key_arr = torch.from_numpy(test_label_key_arr)
            test_label_key_arr = test_label_key_arr.to(self.__ctx).float()
            test_label_arr = torch.one_hot(test_label_key_arr, self.__label_n)

            target_row_arr = np.arange(self.__target_observed_arr.shape[0])
            np.random.shuffle(target_row_arr)

            target_batch_arr = self.__target_observed_arr[target_row_arr[:self.batch_size]]
            target_batch_arr = torch.from_numpy(target_batch_arr)
            target_batch_arr = target_batch_arr.to(self.__ctx).float()
            target_batch_arr = self.pre_normalize(target_batch_arr)

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)
                target_batch_arr = self.__noiseable_data.noise(target_batch_arr)

            train_encoded_observed_arr = training_batch_arr
            train_decoded_observed_arr = training_batch_arr
            train_encoded_masked_arr = None
            train_decoded_masked_arr = None
            test_encoded_observed_arr = test_batch_arr
            test_decoded_observed_arr = test_batch_arr
            test_encoded_masked_arr = None
            test_decoded_masked_arr = None

            list_tuple = self.create_pretext_task_samples_list(target_batch_arr)
            (
                pretext_encoded_observed_arr_list, 
                pretext_decoded_observed_arr_list, 
                pretext_encoded_mask_arr_list,
                pretext_decoded_mask_arr_list,
                pretext_label_arr_list
            ) = list_tuple

            yield (
                train_encoded_observed_arr,
                train_decoded_observed_arr,
                train_encoded_masked_arr,
                train_decoded_masked_arr,
                test_encoded_observed_arr,
                test_decoded_observed_arr,
                test_encoded_masked_arr,
                test_decoded_masked_arr,
                training_label_arr,
                test_label_arr,
                pretext_encoded_observed_arr_list, 
                pretext_decoded_observed_arr_list, 
                pretext_encoded_mask_arr_list,
                pretext_decoded_mask_arr_list,
                pretext_label_arr_list
            )

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if torch.max(arr) != torch.min(arr):
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - torch.min(arr)) / (torch.max(arr) - torch.min(arr) + n)
        elif self.norm_mode == "z_score":
            std = torch.std(arr)
            if std == 0:
                std += 1e-08
            arr = (arr - torch.mean(arr)) / std

        arr = arr * self.scale
        return arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        for i in range(1, self.__test_observed_arr.shape[0] // self.batch_size):
            test_batch_arr = self.__test_observed_arr[(i-1)*self.batch_size:i*self.batch_size]
            yield (
                None,
                None,
                None,
                None,
                test_batch_arr,
                test_batch_arr,
                None,
                None,
                None,
                None,
                None, 
                None, 
                None,
                None,
                None
            )



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from logging import getLogger
import os

import torch

from accelbrainbase.extractabledata.image_extractor import ImageExtractor
from accelbrainbase.iteratabledata.drcn_iterator import DRCNIterator as _DRCNIterator
from accelbrainbase.noiseable_data import NoiseableData


class DRCNIterator(_DRCNIterator):
    '''
    Iterator that draws from image files and generates `mxnet.ndarray`.

    References:
        - Ghifary, M., Kleijn, W. B., Zhang, M., Balduzzi, D., & Li, W. (2016, October). Deep reconstruction-classification networks for unsupervised domain adaptation. In European Conference on Computer Vision (pp. 597-613). Springer, Cham.

    '''

    def __init__(
        self,
        image_extractor,
        dir_list,
        target_domain_dir_list,
        test_dir_list=None,
        epochs=300,
        batch_size=20,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        ctx="ctx"
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            dir_list:                       `list` of directories that store your image files in training.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            target_domain_dir_list:         `list` of directories that store your image files in DRCNetwork's domain adaptation.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            test_dir_list:                  `list` of directories that store your image files in test.
                                            If `None`, this value will be equivalent to `dir_list`.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            epochs:                         `int` of epochs of Mini-batch.
            batch_size:                     `int` of batch size of Mini-batch.
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
        '''

        if isinstance(image_extractor, ImageExtractor) is False:
            raise TypeError("The type of `image_extractor` must be `ImageExtractor`.")
        if isinstance(dir_list, list) is False:
            raise TypeError("The type of `dir_list` must be `list`.")
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        dir_list.sort()
        self.__training_file_path_list = [None] * len(dir_list)
        dataset_size = 0
        for i in range(len(dir_list)):
            file_path_list = [dir_list[i] + "/" + file_name for file_name in os.listdir(dir_list[i] + "/")]
            file_path_list.sort()
            self.__training_file_path_list[i] = file_path_list
            dataset_size = dataset_size + len(file_path_list)

        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        target_domain_dir_list.sort()
        self.__target_domain_file_path_list = [None] * len(target_domain_dir_list)
        for i in range(len(target_domain_dir_list)):
            file_path_list = [target_domain_dir_list[i] + "/" + file_name for file_name in os.listdir(target_domain_dir_list[i] + "/")]
            file_path_list.sort()
            self.__target_domain_file_path_list[i] = file_path_list

        if test_dir_list is not None and isinstance(test_dir_list, list) is False:
            raise TypeError("The type of `test_dir_list` must be `list`.")
        elif test_dir_list is None:
            test_dir_list = dir_list
            self.__test_file_path_list = self.__training_file_path_list
        else:
            test_dir_list.sort()
            self.__test_file_path_list = [None] * len(test_dir_list)
            for i in range(len(test_dir_list)):
                file_path_list = [test_dir_list[i] + "/" + file_name for file_name in os.listdir(test_dir_list[i] + "/")]
                file_path_list.sort()
                self.__test_file_path_list[i] = file_path_list

        self.__image_extractor = image_extractor
        self.__dir_list = dir_list
        self.__test_dir_list = test_dir_list
        self.__target_domain_dir_list = target_domain_dir_list

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data
        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
            - `mxnet.ndarray` of obsrved data points in target domain.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None
            training_label_arr, test_label_arr = None, None
            target_domain_batch_arr = None
            for batch_size in range(self.batch_size):
                dir_key = np.random.randint(low=0, high=len(self.__training_file_path_list))

                training_one_hot_arr = torch.zeros((1, len(self.__training_file_path_list)), device=self.__ctx)
                training_one_hot_arr[0, dir_key] = 1

                file_key = np.random.randint(low=0, high=len(self.__training_file_path_list[dir_key]))
                training_data_arr = self.__image_extractor.extract(
                    path=self.__training_file_path_list[dir_key][file_key],
                )
                training_data_arr = self.pre_normalize(training_data_arr)

                test_dir_key = np.random.randint(low=0, high=len(self.__test_file_path_list))

                test_one_hot_arr = torch.zeros((1, len(self.__test_file_path_list)), device=self.__ctx)
                test_one_hot_arr[0, test_dir_key] = 1

                file_key = np.random.randint(low=0, high=len(self.__test_file_path_list[test_dir_key]))
                test_data_arr = self.__image_extractor.extract(
                    path=self.__test_file_path_list[test_dir_key][file_key],
                )
                test_data_arr = self.pre_normalize(test_data_arr)

                target_domain_dir_key = np.random.randint(low=0, high=len(self.__target_domain_file_path_list))

                target_domain_one_hot_arr = torch.zeros((1, len(self.__target_domain_file_path_list)), device=self.__ctx)
                target_domain_one_hot_arr[0, target_domain_dir_key] = 1

                target_domain_file_key = np.random.randint(low=0, high=len(self.__target_domain_file_path_list[target_domain_dir_key]))
                target_domain_data_arr = self.__image_extractor.extract(
                    path=self.__target_domain_file_path_list[target_domain_dir_key][target_domain_file_key],
                )
                target_domain_data_arr = self.pre_normalize(target_domain_data_arr)

                training_data_arr = torch.unsqueeze(training_data_arr, dim=0)
                test_data_arr = torch.unsqueeze(test_data_arr, dim=0)
                target_domain_data_arr = torch.unsqueeze(target_domain_data_arr, dim=0)

                if training_batch_arr is not None:
                    training_batch_arr = torch.cat((training_batch_arr, training_data_arr), dim=0)
                else:
                    training_batch_arr = training_data_arr
                
                if test_batch_arr is not None:
                    test_batch_arr = torch.cat((test_batch_arr, test_data_arr), dim=0)
                else:
                    test_batch_arr = test_data_arr

                if training_label_arr is not None:
                    training_label_arr = torch.cat((training_label_arr, training_one_hot_arr), dim=0)
                else:
                    training_label_arr = training_one_hot_arr

                if test_label_arr is not None:
                    test_label_arr = torch.cat((test_label_arr, test_one_hot_arr), dim=0)
                else:
                    test_label_arr = test_one_hot_arr

                if target_domain_batch_arr is not None:
                    target_domain_batch_arr = torch.cat((target_domain_batch_arr, target_domain_data_arr), dim=0)
                else:
                    target_domain_batch_arr = target_domain_data_arr

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)
                target_domain_batch_arr = self.__noiseable_data.noise(target_domain_batch_arr)

            yield training_batch_arr, training_label_arr, test_batch_arr, test_label_arr, target_domain_batch_arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        scan_file_path_list = []
        for dir_key in range(len(self.__test_file_path_list)):
            for file_key in range(len(self.__test_file_path_list[dir_key])):
                file_path = self.__test_file_path_list[dir_key][file_key]
                scan_file_path_list.append(file_path)

        import random
        random.shuffle(scan_file_path_list)

        test_batch_arr = None
        file_path_list = []
        for file_path in scan_file_path_list:
            test_data_arr = self.__image_extractor.extract(
                path=file_path,
            )
            test_data_arr = self.pre_normalize(test_data_arr)

            test_data_arr = torch.unsqueeze(test_data_arr, dim=0)
            if test_batch_arr is not None:
                test_batch_arr = torch.cat((test_batch_arr, test_data_arr), dim=0)
            else:
                test_batch_arr = test_data_arr

            file_path_list.append(file_path)

            if test_batch_arr.shape[0] == self.batch_size:
                test_batch_arr_ = test_batch_arr
                test_batch_arr = None
                _file_path_list = file_path_list
                file_path_list = []
                yield None, None, test_batch_arr_, _file_path_list
            elif test_batch_arr.shape[0] > self.batch_size:
                raise ValueError()

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if arr.max() != arr.min():
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - arr.min()) / (arr.max() - arr.min() + n)
        elif self.norm_mode == "z_score":
            std = arr.std()
            if std == 0:
                std += 1e-08
            arr = (arr - arr.mean()) / std

        arr = arr * self.scale
        return arr



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.neuralnetworks.auto_encoder import AutoEncoder
from accelbrainbase.observabledata._torch.neural_networks import NeuralNetworks
from accelbrainbase.iteratable_data import IteratableData
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer, required
from torch.optim.adam import Adam
from logging import getLogger


class VariationalAutoEncoder(AutoEncoder):
    """
    Variational Auto-Encoder.

    References:
        - Kingma, D. P., & Welling, M. (2014). Auto-encoding variational Bayes, May 2014. arXiv preprint arXiv:1312.6114.
    """

    def __init__(
        self,
        encoder,
        decoder,
        computable_loss,
        learning_rate=1e-05,
        ctx="cpu",
        tied_weights_flag=False,
        not_init_flag=False,
        hidden_dim=None,
    ):
        '''
        Init.

        Args:
            encoder:                        is-a `NeuralNetworks`.
            decoder:                        is-a `NeuralNetworks`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            learning_rate:                  `float` of learning rate.
            observed_activation:            `act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` 
                                            that activates observed data points.

            ctx:                            Context-manager that changes the selected device.
            tied_weights_flag:              `bool` of flag to tied weights or not.
            not_init_flag:                  `bool` of whether initialize parameters or not.
        '''
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            computable_loss=computable_loss,
            learning_rate=learning_rate,
            ctx=ctx,
            tied_weights_flag=tied_weights_flag,
            not_init_flag=not_init_flag,
        )
        self.__ctx = ctx
        self.__learning_rate = learning_rate
        self.__not_init_flag = not_init_flag
        self.__hidden_dim = hidden_dim
        self.__computable_loss = computable_loss
        logger = getLogger("accelbrainbase")
        self.__logger = logger

    def parameters(self):
        '''
        '''
        return [
            {
                "params": self.encoder.parameters(),
            },
            {
                "params": self.decoder.parameters(),
            },
            {
                "params": self.encoder_mu.parameters(),
            },
            {
                "params": self.encoder_var.parameters(),
            }
        ]

    def initialize_params(self, input_dim):
        '''
        Initialize params.
        '''
        if self.encoder_optimizer is not None and self.decoder_optimizer is not None:
            return

        self.__encoder_input_dim = input_dim
        self.__decoder_input_dim = self.__hidden_dim

        self.encoder_mu = nn.Linear(
            self.encoder.units_list[-1],
            self.__hidden_dim
        )
        self.encoder_var = nn.Linear(
            self.encoder.units_list[-1],
            self.__hidden_dim
        )
        self.to(self.__ctx)

        if self.initializer_f is None:
            self.encoder_mu.weight = torch.nn.init.xavier_normal_(
                self.encoder_mu.weight,
                gain=1.0
            )
            self.encoder_var.weight = torch.nn.init.xavier_normal_(
                self.encoder_var.weight,
                gain=1.0
            )
        else:
            self.encoder_mu.weight = self.initializer_f(self.encoder_mu.weight)
            self.encoder_var.weight = self.initializer_f(self.encoder_var.weight)

        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                self.encoder.initialize_params(self.__encoder_input_dim)
                self.decoder.initialize_params(self.__decoder_input_dim)
                self.encoder_optimizer = self.encoder.optimizer
                self.decoder_optimizer = self.decoder.optimizer

                if self.optimizer_f is None:
                    self.encoder_mu_optimizer = Adam(
                        self.encoder_mu.parameters(), 
                        lr=self.__learning_rate,
                    )
                    self.encoder_var_optimizer = Adam(
                        self.encoder_var.parameters(), 
                        lr=self.__learning_rate,
                    )
                else:
                    self.encoder_mu_optimizer = self.optimizer_f(
                        self.encoder_mu_optimizer.parameters(), 
                    )
                    self.encoder_var_optimizer = self.optimizer_f(
                        self.encoder_var_optimizer.parameters(), 
                    )

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `IteratableData`.
        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                batch_size = batch_observed_arr.shape[0]
                self.initialize_params(
                    input_dim=batch_observed_arr.reshape(batch_size, -1).shape[1]
                )
                self.encoder_optimizer.zero_grad()
                self.encoder_mu_optimizer.zero_grad()
                self.encoder_var_optimizer.zero_grad()
                self.decoder_optimizer.zero_grad()
                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                self.encoder_optimizer.step()
                self.encoder_mu_optimizer.step()
                self.encoder_var_optimizer.step()
                self.decoder_optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)

                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )
                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()
                    self.__loss_list.append((_loss, _test_loss))
                    self.__logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.epoch = epoch
        self.__logger.debug("end. ")

    def inference(self, observed_arr):
        '''
        Inference the feature points.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `tensor` of inferenced feature points.
        '''
        pred_arr = self.forward(observed_arr)
        return pred_arr

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `tensor`.
            labeled_arr:    `tensor`.

        Returns:
            loss.
        '''
        labeled_arr = pred_arr.reshape_as(pred_arr)
        reconstruction_loss = self.__computable_loss(pred_arr, labeled_arr)
        kl_loss = -0.5 * torch.sum(1 + self.__log_var_arr - self.__mu_arr.pow(2) - self.__log_var_arr.exp())
        loss = reconstruction_loss + kl_loss
        return loss

    def forward(self, x):
        '''
        Forward with torch.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        batch_size = x.shape[0]
        self.initialize_params(
            input_dim=x.reshape(batch_size, -1).shape[1]
        )

        encoded_arr = self.encoder.inference(x)
        mu_arr = self.encoder_mu(encoded_arr)
        log_var_arr = self.encoder_var(encoded_arr)
        eps_arr = torch.randn_like(torch.exp(log_var_arr))
        Z_arr = mu_arr + torch.exp(log_var_arr / 2) * eps_arr

        self.feature_points_arr = Z_arr
        self.__mu_arr = mu_arr
        self.__log_var_arr = log_var_arr
        decoded_arr = self.decoder.inference(Z_arr)
        self.__pred_arr = decoded_arr
        return decoded_arr

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        encoder_filename = filename.replace("." + _format, "_encoder." + _format)
        decoder_filename = filename.replace("." + _format, "_decoder." + _format)
        return encoder_filename, decoder_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.save_parameters(encoder_filename)
        self.decoder.save_parameters(decoder_filename)
        torch.save(
            {
                'epoch': self.epoch,
                'loss': self.loss_arr,
                'encoder_mu_state_dict': self.encoder_mu.state_dict(),
                'encoder_var_state_dict': self.encoder_var.state_dict(),
                'encoder_mu_optimizer_state_dict': self.encoder_mu_optimizer.state_dict(),
                'encoder_var_optimizer_state_dict': self.encoder_var_optimizer.state_dict(),
            }, 
            filename
        )

    def load_parameters(self, filename, ctx=None, strict=True):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            Context-manager that changes the selected device.
            strict:         Whether to strictly enforce that the keys in state_dict match the keys returned by this module’s state_dict() function. Default: `True`.
        '''
        encoder_filename, decoder_filename = self.__rename_file(filename)
        self.encoder.load_parameters(encoder_filename, ctx=ctx, strict=strict)
        self.decoder.load_parameters(decoder_filename, ctx=ctx, strict=strict)

        checkpoint = torch.load(filename)
        self.epoch = checkpoint['epoch']
        self.__loss_list = checkpoint['loss'].tolist()

        self.encoder_mu.load_state_dict(checkpoint['encoder_mu_state_dict'], strict=strict)
        self.encoder_var.load_state_dict(checkpoint['encoder_var_state_dict'], strict=strict)
        self.encoder_mu_optimizer.load_state_dict(
            checkpoint['encoder_mu_optimizer_state_dict']
        )
        self.encoder_var_optimizer.load_state_dict(
            checkpoint['encoder_var_optimizer_state_dict']
        )

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)




# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.adversarialmodel.generative_model import GenerativeModel
from accelbrainbase.samplabledata.condition_sampler import ConditionSampler
from accelbrainbase.samplabledata.noise_sampler import NoiseSampler
from accelbrainbase.observabledata._torch.convolutionalneuralnetworks.convolutional_auto_encoder import ConvolutionalAutoEncoder as _ConvolutionalAutoEncoder

import numpy as np
import torch

from logging import getLogger


class ConvolutionalAutoEncoder(GenerativeModel):
    '''
    Generative model, which draws samples from the fake distribution, 
    in the Generative Adversarial Networks(GANs).

    The Generative Adversarial Networks(GANs) (Goodfellow et al., 2014) framework establishes 
    a min-max adversarial game between two neural networks – a generative model, G, and a 
    discriminative model, D. The discriminator model, D(x), is a neural network that computes 
    the probability that a observed data point x in data space is a sample from the data 
    distribution (positive samples) that we are trying to model, rather than a sample from our 
    generative model (negative samples). 
    
    Concurrently, the generator uses a function G(z) that maps samples z from the prior p(z) to 
    the data space. G(z) is trained to maximally confuse the discriminator into believing that 
    samples it generates come from the data distribution. The generator is trained by leveraging 
    the gradient of D(x) w.r.t. x, and using that to modify its parameters.

    References:
        - Gauthier, J. (2014). Conditional generative adversarial nets for convolutional face generation. Class Project for Stanford CS231N: Convolutional Neural Networks for Visual Recognition, Winter semester, 2014(5), 2.
        - Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., ... & Bengio, Y. (2014). Generative adversarial nets. In Advances in neural information processing systems (pp. 2672-2680).
        - Makhzani, A., Shlens, J., Jaitly, N., Goodfellow, I., & Frey, B. (2015). Adversarial autoencoders. arXiv preprint arXiv:1511.05644.
        - Mirza, M., & Osindero, S. (2014). Conditional generative adversarial nets. arXiv preprint arXiv:1411.1784.
        - Salimans, T., Goodfellow, I., Zaremba, W., Cheung, V., Radford, A., & Chen, X. (2016). Improved techniques for training gans. In Advances in neural information processing systems (pp. 2234-2242).
        - Zhao, J., Mathieu, M., & LeCun, Y. (2016). Energy-based generative adversarial network. arXiv preprint arXiv:1609.03126.
        - Warde-Farley, D., & Bengio, Y. (2016). Improving generative adversarial networks with denoising feature matching.

    '''

    def __init__(
        self, 
        noise_sampler, 
        model, 
        optimizer_f=None,
        condition_sampler=None,
        conditonal_dim=1,
        learning_rate=1e-05,
        ctx="cpu", 
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            noise_sampler:                  is-a `NoiseSampler`.
            model:                          is-a `mxnet.gluon.hybrid.hybridblock.HybridBlock`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            condition_sampler:              is-a `ConditionSampler` of sampler to draw conditons from user-defined distributions.
            conditonal_dim:                 `int` of the dimension to be concated conditions and observed data points.
            learning_rate:                  `float` of learning rate.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            ctx:                            `mx.cpu()` or `mx.gpu()`.

        '''
        if isinstance(model, _ConvolutionalAutoEncoder) is False:
            raise TypeError("The type of `model` must be `accelbrainbase.observabledata._torch.convolutionalneuralnetworks.convolutional_auto_encoder.ConvolutionalAutoEncoder`.")

        super().__init__(
            noise_sampler=noise_sampler, 
            model=model, 
            optimizer_f=optimizer_f,
            condition_sampler=condition_sampler,
            conditonal_dim=conditonal_dim,
            learning_rate=learning_rate,
            ctx=ctx, 
            not_init_flag=not_init_flag,
        )

        self.condition_sampler = condition_sampler
        self.conditonal_dim = conditonal_dim

    def draw(self):
        '''
        Draw samples from the fake distribution.

        Returns:
            `Tuple` of `mxnet.ndarray`s.
        '''
        if self.condition_sampler is None:
            observed_arr = self.noise_sampler.draw()
            encoded_arr = self.model.encoder(observed_arr)
            decoded_arr = self.model.decoder(encoded_arr)
            generated_arr = encoded_arr
        else:
            condition_arr, sampled_arr = self.condition_sampler.draw()
            if sampled_arr is not None:
                sampled_arr = sampled_arr + self.noise_sampler.draw()
                encoded_arr = self.model.encoder(sampled_arr)
                decoded_arr = self.model.decoder(encoded_arr)

                observed_arr = sampled_arr

                generated_arr = torch.cat(
                    (
                        encoded_arr,
                        condition_arr
                    ),
                    dim=self.conditonal_dim
                )
            else:
                condition_arr = condition_arr + self.noise_sampler.draw()
                encoded_arr = self.model.encoder(condition_arr)
                decoded_arr = self.model.decoder(encoded_arr)

                observed_arr = condition_arr
                generated_arr = encoded_arr

        return generated_arr, observed_arr, decoded_arr
